#include "lotromanager.h"
#include "app.h"
#include "filesystem.h"

#include <QStringList>

LotroManager::LotroManager(QObject *parent) : QObject(parent)
{
    dat_files_.resize(5);
}

void LotroManager::saveDatFiles() {
    for (auto dat_file : dat_files_)
        dat_file.CloseDatFile();
    App *app = &App::getInstance();
    app->helper->setState("free");
}

void LotroManager::startGame(){
    App *app = &App::getInstance();

    for (auto &dat_file : dat_files_)
        dat_file.CloseDatFile();

    QStringList args;
    args << "-skiprawdownload" << "-nosplash";
    if (dat_locale_ == PATCHED)
        args << "-disablePatch";


    if(FileSystem::fileExists(QApplication::applicationDirPath() + "/user.ini")){
        QSettings login(QApplication::applicationDirPath() + "/user.ini", QSettings::IniFormat );
        login.beginGroup("Account");
        QString username = login.value("username", "").toString();
        QString password = login.value("password", "").toString();
        login.endGroup();
        args << "-username" << username << "-password" << password;
    }

    qDebug() << "Запускаем игру со следующими аргументами: " << args;

    QFile f(app->config->getValue("Local", "folder") + "/TurbineLauncher.exe");
    QProcess process;

    if (FileSystem::fileExists(f.fileName())) {
        if(f.fileName().contains(" ")) f.setFileName("\"" + f.fileName() + "\"");
        process.startDetached(f.fileName(), args);
        process.waitForFinished(-1);
        process.deleteLater();
        QApplication::quit();
    } else {
        QString text = "Запуск не удался";
        QString info = "Не удалось запустить игру. Не найден файл TurbineLauncher.exe или невозможно получить к нему доступ. "
                       "Проверьте в 'Настройках', что у вас указан верный путь к игре, закрыт лаунчер игры и повторите попытку запуска.";
        app->helper->myDialogBox(text, info, "OK", "Отмена", "gandalf.png", "", "", 400, 160, true, false);
    }
}

bool LotroManager::openDatFile(int id) {
    App *app = &App::getInstance();
    QStringList dat_files_name_list;

    dat_files_name_list << app->config->getValue("Local", "file")
                        << "client_general.dat"
                        << "client_sound.dat"
                        << "client_surface.dat"
                        << "client_highres.dat";

    if (dat_files_name_list[id] != nullptr) {

        qDebug() << dat_files_name_list[id];

        QString filename = app->config->getValue("Local", "folder") + "/" + dat_files_name_list[id];
        QString execpath = app->config->getValue("Local", "folder") + "/" + "TurbineLauncher.exe";

        if (!FileSystem::fileExists(filename) && FileSystem::fileExists(execpath)) {
            qWarning("%s:%i: %s\n", __FILE__, __LINE__, "Файл данных не найден, но папка верная");
            QString text = "Не найден файл ресурсов";
            QString msg = "Файла ресурсов " + filename + ", похоже, не существует. Запустить лаунчер игры для "
                                                         "начала автоматического скачивания?";
            app->helper->myDialogBox(text, msg, "OK", "Отмена", "gandalf.png", "loaddatfile", "", 400, 180, true, true);
            app->helper->setState("free");
            return false;
        }

        if (!FileSystem::fileExists(filename) && !FileSystem::fileExists(execpath)) {
            QString text = "Не найдены файлы игры!";
            QString msg = "Похоже, папка с игрой указана неверно. Укажите, пожалуйста, путь к папке с игрой. "
                          "Опознать папку с игрой можно по наличию папок 3rdparty, backup, bindat, de, en, forward и т.п.";
            app->helper->myDialogBox(text, msg, "OK", "Выход", "gandalf.png", "choosedatpath", "exit", 400, 170, true, true);
            app->helper->setState("free");
            return false;
        }

        qDebug("%s:%i: %s%s", __FILE__, __LINE__, "Инициализация dat-файла. Открываем файл ", filename.toLocal8Bit().toStdString().c_str());
        int dat_state = dat_files_[id].InitDatFile(filename.toLocal8Bit().toStdString(), id);
        qDebug("%s:%i: %s%d", __FILE__, __LINE__, "Состояние dat-файла: ", dat_state);

        if(dat_state > 0){
            app->helper->setState("busy");
            return true;
        } else {
            QString text = "Нет доступа к  файлам игры";
            QString info = "Возможно в данный момент запущены клиент или лаунчер игры либо попытка доступа блокируется настройкой прав, антивирусом или другими приложениями. Завершите все процессы, использующие файлы игры и повторите попытку.";
            app->helper->myDialogBox(text, info, "OK", "Отмена", "gandalf.png", "", "", 400, 200, true, false);
            app->helper->setState("free");
            return false;
        }
    }
    return true;
}

void LotroManager::applyPatch(QString name){
    App *app = &App::getInstance();
    app->window->ui->hintLabel->setText("Установка патчей");
    app->window->ui->exthintLabel->setText("Ожидайте. Идет установка выбранных патчей");
    qDebug("%s:%i: %s%s", __FILE__, __LINE__, "Начинаем применение патча ", app->helper->stringToChar(name));
    QLabel* label = app->window->findChild<QLabel*>(name + "Status");
    QDir dir(QApplication::applicationDirPath() + "/data");
    if (!dir.exists()) {
        app->window->ui->hintLabel->setText("Установка патчей завершена с ошибкой");
        app->window->ui->exthintLabel->setText("Ошибка - не найдена папка data");
        return;
    }

    QStringList paths = dir.entryList(QStringList(name + "*"));
    qDebug() << app->helper->stringToChar("data/" + paths.first());
    if(!paths.empty()){
       database_.InitDatabase(app->helper->stringToChar("data/" + paths.first()));
        int indb = database_.CountRows();
        qInfo("%s:%i: %s%d", __FILE__, __LINE__, "Файлов в обновлении: ", indb);
        for(int i = 0; i<=indb; i++){
            processFile();
            if (label != nullptr)
                label->setText(QString("Применение ... ") + QString::fromLatin1("%1").arg(i*100.0/indb, 3, 'f', 1) + QString("%"));
        }
        if(label != nullptr)label->setText("Готово");
        //for (int i = 0; i < 5; i++) app->datfiles[i].CommitChanges();
        app->config->setValue("Applied", name, paths.first());
        database_.CloseDatabase();
    }
    app->window->ui->hintLabel->setText("Установка патчей завершена");
    app->window->ui->exthintLabel->setText("Применение выбранных вами патчей закончено");
}

void LotroManager::applyGlobal(){
    App *app = &App::getInstance();

    // Применяем патч с заставочными экранами
    if(app->config->getValue("Editor", "screens") == "true"){
        applyLoadscreens();
    } else {
        QString dateline = app->config->getValue("Datetime", "loadscreens");
        if(dateline != "-1"){
            QStringList date = dateline.split(" ");
            app->window->ui->loadscreensStatus->setText(date[0]);
        }
    }

    // Применяем остальные патчи
    QStringList names;
    QStringList need;
    QFuture<void> f;
    f = QtConcurrent::run([=]() {});
    names  << "fonts" << "sounds" << "texts" << "images" << "videos" << "textures";
    QDir dir(QApplication::applicationDirPath() + "/data");

    foreach(QString name, names) {

        QStringList list = dir.entryList(QStringList(name + "*"));

        if (list.size() > 0 && list.first() != ""){
            QString isset = app->config->getValue("Applied", name);
            if (list.first() != isset && app->config->getValue("Editor", name) == "true") {
                need.append(name);
            }
        }
    }

    Helper *helper = new Helper();
    QFutureWatcher<void> * watcher = new QFutureWatcher<void>();
    connect(watcher, SIGNAL(finished()), helper, SLOT(handleFinisheddd()));

    if(need.size() > 0){
        watcher->setFuture(QtConcurrent::run([=]() {
            if(openDatFile(0)){
                foreach(QString name, need) {
                    applyPatch(name);
                }
                saveDatFiles();
            }
        }));
    } else {
        installMicroPatch();
    }

}

bool LotroManager::isDatReady(){
    App *app = &App::getInstance();
    bool free;
    QString dir = app->config->getValue("Local", "folder");
    QString file = app->config->getValue("Local", "file");
    QFile fl(dir+"/" + file);
    qDebug() << fl.fileName();
    if(FileSystem::fileExists(fl.fileName())){
        QDir game_dir(dir);
        free = game_dir.rename(fl.fileName(), fl.fileName() + "99");
        game_dir.rename(fl.fileName() + "99", fl.fileName());
    } else {
        free = false;
    }
    if(free == true) qDebug() << "Dat is FREE"; else qDebug() << "Dat is BUSY";
    return free;
}

int LotroManager::checkDatFile(){
    //bool CheckIfUpdatedByGame(); // Обновлялся ли игрой
    //bool CheckIfNotPatched(); // Был ли хоть раз пропатчен новым лаунчером
    //bool CheckIfPatchedByOldLauncher(); // Был ли хоть раз пропатчен старым лаунчером

    App *app = &App::getInstance();
    if (!openDatFile(0)) {
        return 0;
    }

    if(dat_files_[0].CheckIfUpdatedByGame()){
        return 1;
        //app->ready = false;
        //QString text = "Файлы локализации изменены";
        //QString info = "В процессе последнего запуска игры были установлены обновления игрового клиента. Это могло затронуть файлы локализации. Желаете провести проверку и исправить поврежденные участки перевода? Вы можете в любой момент сделать это самостоятельно, нажав кнопку «Применить патчи заново».";
        //app->helper->myDialogBox(text, info, "OK", "Отмена", "gandalf.png", "applypaths", "", 400, 220, true, true);
    }

    if(dat_files_[0].CheckIfNotPatched() && app->config->getValue("Local", "runfirst") == "1"){
        return 2;
        //app->ready = false;
        //QString text = "Файлы локализации отсутствуют";
        //QString info = "Текущий файл локализации не содержит патчей. Возможно это первый запуск «Наследия» с данным игровым клиентом, или клиент был обновлён игровым лаунчером. Применить выбранные патчи в данному файлу?";
        //app->helper->myDialogBox(text, info, "OK", "Отмена", "gandalf.png", "applypaths", "", 420, 200, true, true);
    }

    if(dat_files_[0].CheckIfPatchedByOldLauncher()){
        return 3;
        //QString text = "Файлы локализации не являются оригинальными";
        //QString info = "Внимание! Присутствующие в вашем игровом клиенте файлы локализации не являются оригинальными, поскольку были модифицированы другой программой.<br/><strong>Мы настоятельно рекомендуем перекачать оригинальные версии файлов, поскольку правильная работа Наследия с данными файлами не гарантируется!</strong>";
        //app->helper->myDialogBox(text, info, "Перекачать", "Не надо", "gandalf.png", "loaddatfile", "", 550, 200, true, true);
    }
    return 0;
}

void LotroManager::applyMicroPatch(){
    return;
    App *app = &App::getInstance();
    app->logSectionStart("Установка мгновенных обновлений");
    app->helper->setState("busy");
    app->window->ui->hintLabel->setText("Установка последних обновлений");
    app->window->ui->exthintLabel->setText("Подождите пока закончится применение обновлений");

    database_.InitDatabase(app->helper->stringToChar(app->network->micropatch));
    int indb = database_.CountRows();
    qInfo("%s:%i: %s%d", __FILE__, __LINE__, "Файлов в обновлении: ", indb);

    for(int i = 0; i <= indb; i++ ) {
        processFile();
    }
    //for (int i = 0; i < 5; i++) app->datfiles[i].CommitChanges();
    database_.CloseDatabase();

    app->config->setValue("Updates", "update", QDate::currentDate().toString("yyyy-MM-dd"));

    app->helper->setState("free");
    qInfo("%s:%i: %s", __FILE__, __LINE__, "Выполнено.");
    app->logSectionEnd();
}

void LotroManager::applyLoadscreens(){
    App *app = &App::getInstance();
    app->logSectionStart("Установка заставочных экранов");
    app->helper->setState("busy");

    QString datafolder = QApplication::applicationDirPath() + "/data";
    QDir dir(datafolder);
    if (!dir.exists())
        return;
    QStringList paths = dir.entryList(QStringList("loadscreens*"));

    if(!paths.empty()){

        if(paths.first() == app->config->getValue("Applied", "screens")){
            app->helper->setState("free");
            return;
        }

        QString lang = app->config->getValue("Local", "lang");

        if (lang == "-1") {
            app->config->setValue("Local", "lang", "en");
            lang = "en";
        }

        QString folder = app->config->getValue("Local", "folder") + "raw/" + lang + "/logo/";
        SubfileData subfile;
        QStringList filenames;
        QString mainscreen = lang == "en" ? "lotro_ad_pregame.jpg" : "lotro_ad_pregame_" + lang + ".jpg";
        filenames << mainscreen << "lotro_generic_teleport_screen_01.jpg" << "lotro_generic_teleport_screen_02.jpg"
                  << "lotro_generic_teleport_screen_03.jpg" << "lotro_generic_teleport_screen_04.jpg" << "lotro_generic_teleport_screen_05.jpg"
                  << "lotro_generic_teleport_screen_06.jpg" << "lotro_generic_teleport_screen_07.jpg" << "lotro_generic_teleport_screen_08.jpg"
                  << "lotro_generic_teleport_screen_09.jpg" << "lotro_generic_teleport_screen_10.jpg";

        QString basename = datafolder + "/" + paths.first();
        database_.InitDatabase(basename.toStdString());
        int indb = database_.CountRows();
        qInfo("%s:%i: %s%d", __FILE__, __LINE__, "Файлы загрузочных экранов: ", indb);
        for(int i = 0; i<indb; i++){
            subfile = database_.GetNextFile();
            qDebug() << i;
            if (!subfile.Empty()){
                qInfo("%s:%i: %s", __FILE__, __LINE__, app->helper->stringToChar(folder + filenames[i]));
                QFile::remove(folder + filenames[i]);
                subfile.binary_data.WriteToFile((folder + filenames[i]).toStdString());
            }
        }
        database_.CloseDatabase();

    }
    app->helper->setState("free");
    app->window->ui->loadscreensStatus->setText("Готово");
    app->config->setValue("Applied", "screens", paths.first());
    qInfo("%s:%i: %s", __FILE__, __LINE__, "Выполнено.");
    app->logSectionEnd();
}

void LotroManager::processFile(){
    App *app = &App::getInstance();
    SubfileData subfile;
    subfile = database_.GetNextFile();
    if (subfile.Empty()){
        qInfo("%s:%i: %s", __FILE__, __LINE__, "Достигнут конец файла.");
    } else {
        int dat_id = subfile.options["did"].as<int>();
        int dat_state = dat_files_[dat_id].InitDatFile(app->config->getDatPath(dat_id), dat_id);
        //qDebug("%s:%i: %s%d", __FILE__, __LINE__, "Состояние dat-файла: ", dat_state);
        if(dat_state > 0) dat_files_[dat_id].PatchFile(subfile);
    }
}

void LotroManager::installMicroPatch(){
    return;
    App *app = &App::getInstance();
    if(app->config->getValue("Updates", "micro") != "true") return;

    app->helper->setState("busy");
    app->window->ui->hintLabel->setText("Установка последних обновлений");
    app->window->ui->exthintLabel->setText("Идет проверка и установка новых обновлений");
    QFuture<void> future = QtConcurrent::run([=]() {
        prepareMicroPatch();
        applyMicroPatch();
        QDir dir("/data/micro");
        FileSystem::clearFolder(dir);
    });
    app->window->ui->hintLabel->setText("Установка завершена");
    app->window->ui->exthintLabel->setText("Обновления успешно установлены");
    app->helper->setState("free");

}

void LotroManager::prepareMicroPatch(){
    return;
    App *app = &App::getInstance();
    app->logSectionStart("Автоматическая проверка обновлений");
    app->helper->setState("busy");
    app->window->ui->hintLabel->setText("Загрузка последних обновлений");
    app->window->ui->exthintLabel->setText("Подождите пока загрузятся последние изменения в системе переводов");

    QString date = app->config->getValue("Updates", "update");
    if (date == QDate::currentDate().toString("yyyy-MM-dd"))
        return;

    if (date == "false" || date == "-1")
        date = QDate::currentDate().toString("yyyy-MM-dd");

    //QTime time = QTime::currentTime();
    //int seconds = QTime(0, 0, 0).secsTo(time);
    //int timestamp = app->helper->dateToTimestamp(date + " " + time.toString("hh:mm:ss"), "yyyy-MM-dd hh:mm:ss") - seconds - 1209600;
    int timestamp = app->helper->dateToTimestamp(date, "yyyy-MM-dd") - 1209600;

    qInfo("%s:%i: %s", __FILE__, __LINE__, "Опция активна. Начинаем загрузку обновлений");
    app->network->micropatch = QApplication::applicationDirPath() + "/data/micro/" + QString::number(timestamp) + ".db";
    if(!FileSystem::fileExists(app->network->micropatch)){
        qInfo("%s:%i: %s%s", __FILE__, __LINE__, "Файл патча: ", app->helper->stringToChar(app->network->micropatch));
        app->network->micropatch = app->network->getMicroPath(timestamp);
    }
    app->window->ui->hintLabel->setText("Загрузка завершена");
    app->window->ui->exthintLabel->setText("Все обновления были успешно загружены и доступны к установке");

    app->helper->setState("free");
    qInfo("%s:%i: %s", __FILE__, __LINE__, "Выполнено.");
    app->logSectionEnd();
}

void LotroManager::saveLocale(int locale, QPushButton * button){
    App *app = &App::getInstance();
    QString file;

    if(locale == 0) file = "client_local_English.dat";
    if(locale == 1) file = "client_local_DE.dat";
    if(locale == 2) file = "client_local_FR.dat";

    if(FileSystem::fileExists(app->config->getValue("Local", "folder") + "/" + file)){
        app->window->ui->enButton->setEnabled(true);
        app->window->ui->deButton->setEnabled(true);
        app->window->ui->frButton->setEnabled(true);
        button->setEnabled(false);
        app->config->setValue("Local", "file", file);
    } else {
        QString text = "Данная локализация отсутствует";
        QString info = "В настоящий момент вы не можете выбрать данную опцию, поскольку файл " + file + " отсутствует в вашем игровом клиенте. Чтобы скачать его, запустите официальный лаунчер, выберите желаемый язык в верхней правой части окна и дождитесь, когда загрузка необходимых файлов завершится.";
        app->helper->myDialogBox(text, info, "OK", "Отмена", "gandalf.png", "", "", 400, 220, true, false);
    }
}

bool LotroManager::setGameLocale(QString locale){
    App *app = &App::getInstance();
    app->window->ui->goHint->show();
    app->helper->setState("busy");
    app->window->ui->goLocal->setEnabled(false);
    app->window->ui->goOriginal->setEnabled(false);
    app->window->ui->goCancel->setEnabled(false);

    if(openDatFile(0)){
        LOCALE loc = dat_files_[0].current_locale();
        dat_locale_ = loc;

        if(loc == PATCHED) qDebug() << "Starting " + locale + " version. Current locale is PATCHED";
        if(loc == ORIGINAL) qDebug() << "Starting " + locale + " version. Current locale is ORIGINAL";
        if(loc != ORIGINAL && loc != PATCHED) qDebug() << "Starting " + locale + " version. Current locale is UNKNOWN";

        if(locale == "RU" && loc != PATCHED){ qDebug()<< "Current locale Original"; dat_files_[0].SetLocale(PATCHED); dat_locale_ = PATCHED;}
        if(locale == "Original" && loc != ORIGINAL){ qDebug()<< "Current locale RU"; dat_files_[0].SetLocale(ORIGINAL); dat_locale_ = ORIGINAL;}
        saveDatFiles();
    }

    app->window->ui->goHint->hide();
    app->window->ui->goProgress->hide();
    app->window->ui->startbox->hide();
    app->window->ui->goLocal->setEnabled(true);
    app->window->ui->goOriginal->setEnabled(true);
    app->window->ui->goCancel->setEnabled(true);
    app->helper->setState("free");
    qDebug() << "Locale changed";

    return true;
}

void LotroManager::handleFinisheddd(){
   installMicroPatch();
}
